/* src/net/ocsp_response.c
 * ============================================================
 * OCSP Responder
 * (c) 2001 by Massimiliano Pala
 * OpenCA released software
 * ============================================================
 */
 
#include "general.h"

#include <openssl/err.h>
#include <openssl/ssl.h>

#include <openssl/txt_db.h>
#include <openssl/x509.h>

#include "support.h"
#include "ocsp_db.h"

#include "ocsp_response.h"
#include "configuration.h"
#include "crl.h"

int make_ocsp_response(OCSP_RESPONSE **resp, OCSP_REQUEST *req,
		OCSPD_CONFIG *conf ) {

	OCSP_CERTID *cid = NULL;
	OCSP_BASICRESP *bs = NULL;

	int i, id_count, ret = 1;

	ASN1_GENERALIZEDTIME *thisupd = NULL;
	ASN1_GENERALIZEDTIME *nextupd = NULL;

	id_count = OCSP_request_onereq_count(req);

	if (id_count <= 0) {
		*resp = OCSP_response_create(
				OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL);
		goto end;
	}

	bs = OCSP_BASICRESP_new();
	if (conf->set_nextUpdate)
		nextupd = X509_gmtime_adj(NULL, conf->nmin * 60 + 
				conf->ndays * 3600 * 24 );

	/* Examine each certificate id in the request */
	for (i = 0; i < id_count; i++) {
		OCSP_ONEREQ *one = NULL;
		ASN1_INTEGER *serial = NULL;
		CA_LIST_ENTRY *ca = NULL;
		X509_REVOKED *entry = NULL;

		one = OCSP_request_onereq_get0(req, i);
		cid = OCSP_onereq_get0_id(one);

		/* Get basic request info */
		OCSP_id_get0_info(NULL, NULL, NULL, &serial, cid);

		if( conf->verbose ) {
			syslog( LOG_INFO, "request for certificate serial %ld",
					ASN1_INTEGER_get( serial ) );
		}

		/* Is this request about our CA? */
		if ((ca = ocspd_CA_ENTRY_find( conf, cid )) == NULL) {
			if (conf->verbose) {
				syslog( LOG_INFO, 
				  "request for non reckognized CA [serial %ld]",
					ASN1_INTEGER_get(serial) );
			}
			OCSP_basic_add1_status(bs, cid,
				V_OCSP_CERTSTATUS_UNKNOWN,0,NULL,
				X509_gmtime_adj(NULL, 0),nextupd);
			continue;
		}

		if( ca->compromised > 0 ) {
			OCSP_basic_add1_status( bs, cid,
				V_OCSP_CERTSTATUS_REVOKED, 0, NULL,
				X509_gmtime_adj(NULL, 0),nextupd);
			continue;
		}

		if( ca->crl_status != CRL_OK ) {
			*resp = OCSP_response_create(
				OCSP_RESPONSE_STATUS_TRYLATER, NULL);
			syslog(LOG_ERR, "SENT TRYLATER (%s)", 
					get_crl_status_info (ca->crl_status));
			goto end;
		}

		/* Get the entry from the CRL data, if NULL then the
		   certificate is not revoked */
		entry = ocspd_X509_REVOKED_find( ca, serial );

		/* Sets thisUpdate field to the value of the loaded CRL */
		thisupd = M_ASN1_TIME_dup(ca->lastUpdate);

		if( entry ) {
			OCSP_SINGLERESP *single = NULL;

			int reason = -1;

			/* If extensions are found, process them */
			if( entry->extensions ) {
				ASN1_ENUMERATED *asn = NULL;

				if( (asn = X509_REVOKED_get_ext_d2i( entry, 
						NID_crl_reason,NULL,NULL )) != NULL ) {
					reason = ASN1_ENUMERATED_get( asn );
					ASN1_ENUMERATED_free( asn );
				}
			}

			single = OCSP_basic_add1_status(bs, cid,
						V_OCSP_CERTSTATUS_REVOKED,
						reason, 
						entry->revocationDate,
						thisupd, nextupd);

			if( !single ) {
				syslog( LOG_ERR, "ERROR::Can not generate"
					" basic response");
			}

			/* Check and add the invalidity date */
			if ( entry->extensions ) {
				void *ext = NULL;

				ext = X509_REVOKED_get_ext_d2i( entry, 
					NID_invalidity_date, NULL, NULL );
				if ( ext != NULL ) {
					OCSP_SINGLERESP_add1_ext_i2d(single, 
						NID_invalidity_date, ext, 0, 0);
				}
			}

			if( conf->verbose )
				syslog( LOG_INFO, "Status for %ld is REVOKED",
						ASN1_INTEGER_get(serial) );
		} else if (ca == NULL ) {
			if( conf->verbose ) {
				syslog( LOG_INFO, "status unknown for %ld (unknown CA)",
					ASN1_INTEGER_get(serial) );
			}
			OCSP_basic_add1_status(bs, cid,
				V_OCSP_CERTSTATUS_UNKNOWN,0,NULL,
				thisupd,nextupd);

		} else {
			if( conf->verbose ) {
				syslog( LOG_INFO, "status VALID for %ld",
					ASN1_INTEGER_get(serial) );
			}
			
			OCSP_basic_add1_status(bs, cid,
				V_OCSP_CERTSTATUS_GOOD, 0, NULL, 
				thisupd, nextupd);
		}
	}

	OCSP_copy_nonce(bs, req);
	OCSP_basic_sign(bs, conf->ocspd_cert, conf->ocspd_pkey, 
			conf->digest, conf->other_certs, conf->flags);

	*resp = OCSP_response_create(OCSP_RESPONSE_STATUS_SUCCESSFUL, bs);

	end:

	if(thisupd) ASN1_GENERALIZEDTIME_free(thisupd);
	if(nextupd) ASN1_GENERALIZEDTIME_free(nextupd);
	if(bs) OCSP_BASICRESP_free(bs);

	return ret;
}

int ocspd_resp_send_socket(int connfd, OCSP_RESPONSE *resp) {
	BIO *out = NULL;

	char http_resp[] =
		"HTTP/1.0 200 OK\r\n"
		"Content-type: application/ocsp-response\r\n"
 		"Content-Transfer-Encoding: Binary\r\n"
		"Content-Length: %d\r\n\r\n";

	if (!connfd) {
		syslog(LOG_ERR, "ERROR::socket fd is 0!");
		return 0;
	}

	if((out = BIO_new_socket(connfd, 0)) == NULL ) {
		syslog(LOG_ERR, "ERROR::socket conversion to BIO");
		return(0);
	}

	BIO_printf(out, http_resp, i2d_OCSP_RESPONSE(resp, NULL));
	i2d_OCSP_RESPONSE_bio(out, resp);
	BIO_flush(out);

	BIO_free(out);

	return 1;
}

